@@ -12,9 +12,7 @@
 # See the License for the specific language governing permissions and
 # limitations under the License.
 import inspect
-import logging
 import os
-import random
 import time
 import warnings
 from typing import List, Optional, Union
@@ -35,9 +33,12 @@ from ..core import (
     entropy_from_logits,
     flatten_dict,
     logprobs_from_logits,
+    masked_mean,
+    masked_var,
+    masked_whiten,
+    set_seed,
     stack_dicts,
     stats_to_np,
-    whiten,
 )
 from ..models import SUPPORTED_ARCHITECTURES, PreTrainedModelWrapper, create_reference_model
 from . import AdaptiveKLController, BaseTrainer, FixedKLController, PPOConfig
@@ -53,7 +54,8 @@ tags:
 
 # {model_name}
 
-This is a [TRL language model](https://github.com/lvwerra/trl) that has been fine-tuned with reinforcement learning to guide the model outputs according to a value, function, or human feedback. The model can be used for text generation.
+This is a [TRL language model](https://github.com/lvwerra/trl) that has been fine-tuned with reinforcement learning to
+ guide the model outputs according to a value, function, or human feedback. The model can be used for text generation.
 
 ## Usage
 
@@ -90,24 +92,32 @@ outputs = model(**inputs, labels=inputs["input_ids"])
 class PPOTrainer(BaseTrainer):
     """
     The PPOTrainer uses Proximal Policy Optimization to optimise language models.
+    Note, this trainer is heavily inspired by the original OpenAI learning to summarize work here:
+    https://github.com/openai/summarize-from-feedback
 
     Attributes:
-        **config** (`PPOConfig`) -- Configuration object for PPOTrainer. Check the documentation of `PPOConfig` for more details.
+        **config** (`PPOConfig`) -- Configuration object for PPOTrainer. Check the documentation of `PPOConfig` for more
+         details.
         **model** (`PreTrainedModelWrapper`) -- Model to be optimized, Hugging Face transformer model with a value head.
             Check the documentation of `PreTrainedModelWrapper` for more details.
-        **ref_model** (`PreTrainedModelWrapper`, *optional*) -- Reference model to be used for KL penalty, Hugging Face transformer model with a casual language modelling head.
-            Check the documentation of `PreTrainedModelWrapper` for more details. If no reference model is provided, the
-            trainer will create a reference model with the same architecture as the model to be optimized with shared layers.
-        **tokenizer** (`Union[PreTrainedTokenizer, PreTrainedTokenizerFast]`) -- Tokenizer to be used for encoding the data. Check the documentation of `transformers.PreTrainedTokenizer` and
+        **ref_model** (`PreTrainedModelWrapper`, *optional*) -- Reference model to be used for KL penalty, Hugging Face
+            transformer model with a casual language modelling head. Check the documentation of `PreTrainedModelWrapper`
+            for more details. If no reference model is provided, the trainer will create a reference model with the same
+             architecture as the model to be optimized with shared layers.
+        **tokenizer** (`Union[PreTrainedTokenizer, PreTrainedTokenizerFast]`) -- Tokenizer to be used for encoding the
+            data. Check the documentation of `transformers.PreTrainedTokenizer` and
             `transformers.PreTrainedTokenizerFast` for more details.
-        **dataset** (Union[`torch.utils.data.Dataset`, `datasets.Dataset`], *optional*) -- PyTorch dataset or Hugging Face dataset. This is used to create a PyTorch dataloader. If no dataset is provided,
-            the dataloader must be created outside the trainer users needs to design their own dataloader and make sure the batch
+        **dataset** (Union[`torch.utils.data.Dataset`, `datasets.Dataset`], *optional*) -- PyTorch dataset or Hugging
+            Face dataset. This is used to create a PyTorch dataloader. If no dataset is provided, the dataloader must be
+             created outside the trainer users needs to design their own dataloader and make sure the batch
             size that is used is the same as the one specified in the configuration object.
-        **optimizer** (`torch.optim.Optimizer`, *optional*) -- Optimizer to be used for training. If no optimizer is provided, the trainer will create an Adam optimizer with
-            the learning rate specified in the configuration object.
-        **data_collator** (DataCollatorForLanguageModeling, *optional*) -- Data collator to be used for training and passed along the dataloader
-        **num_shared_layers** (int, *optional*) -- Number of layers to be shared between the model and the reference model, if no reference model is passed. If no number is provided, all the layers
-            will be shared.
+        **optimizer** (`torch.optim.Optimizer`, *optional*) -- Optimizer to be used for training. If no optimizer is
+            provided, the trainer will create an Adam optimizer with the learning rate specified in the configuration
+            object.
+        **data_collator** (DataCollatorForLanguageModeling, *optional*) -- Data collator to be used for training and
+            passed along the dataloader
+        **num_shared_layers** (int, *optional*) -- Number of layers to be shared between the model and the reference
+            model, if no reference model is passed. If no number is provided, all the layers will be shared.
         **lr_scheduler** (`torch.optim.lr_scheduler`, *optional*) -- Learning rate scheduler to be used for training.
     """
 
@@ -150,6 +160,10 @@ class PPOTrainer(BaseTrainer):
                 Learning rate scheduler used for training.
         """
         super().__init__(config)
+
+        # initial seed for reproducible experiments
+        set_seed(config.seed)
+
         # Step 0: check positional arguments validity
         if not isinstance(config, PPOConfig):
             raise ValueError(f"config must be a PPOConfig, got {type(config)}")
@@ -172,14 +186,16 @@ class PPOTrainer(BaseTrainer):
             self.ref_model = ref_model
             if num_shared_layers is not None:
                 warnings.warn(
-                    "num_shared_layers is ignored when ref_model is provided. Two different models are used for the model and the reference model and no layers are shared.",
+                    "num_shared_layers is ignored when ref_model is provided. Two different models are used for the "
+                    "model and the reference model and no layers are shared.",
                     UserWarning,
                 )
         elif ref_model is None:
             self.ref_model = create_reference_model(self.model, num_shared_layers=num_shared_layers)
         else:
             raise ValueError(
-                f"ref_model must be a PreTrainedModelWrapper or `None`, got {type(ref_model)} - supported architectures are: {SUPPORTED_ARCHITECTURES}"
+                f"ref_model must be a PreTrainedModelWrapper or `None`, got {type(ref_model)} - supported "
+                f"architectures are: {SUPPORTED_ARCHITECTURES} "
             )
 
         if not (isinstance(tokenizer, PreTrainedTokenizer) or isinstance(tokenizer, PreTrainedTokenizerFast)):
@@ -247,19 +263,6 @@ class PPOTrainer(BaseTrainer):
         # init the current step
         self.current_step = 0
 
-        # init wandb on the main process:
-        if self.accelerator.is_main_process and self.config.log_with == "wandb":
-            import wandb
-
-            wandb.watch(self.model, log="all")
-
-        if self.config.forward_batch_size > 1 and (self.is_encoder_decoder or self.tokenizer.padding_side == "left"):
-            # warn users that this is not well supported yet
-            logging.warning(
-                "Forward batch size > 1 is not well supported yet for encoder-decoder models and when using `tokenizer.padding_side='left'`. This can lead to unexpected behaviour."
-                " therefore, we recommend using forward_batch_size=1."
-            )
-
     def _filter_kwargs(self, kwargs, target_func):
         """
         filter the keyword arguments that are supported by the target function.
@@ -293,6 +296,7 @@ class PPOTrainer(BaseTrainer):
             batch_size=self.config.batch_size,
             collate_fn=data_collator,
             shuffle=True,
+            drop_last=True,
         )
         return dataloader
 
@@ -332,7 +336,7 @@ class PPOTrainer(BaseTrainer):
         Args:
             query_tensor (`torch.LongTensor`):
                 A tensor of shape (`batch_size`, `seq_len`) containing query tokens.
-            gen_kwargs (dict[str, Any]):
+            generation_kwargs (dict[str, Any]):
                 Keyword arguments for generation.
 
         Returns:
@@ -419,27 +423,64 @@ class PPOTrainer(BaseTrainer):
 
         t = time.time()
 
-        logprobs, ref_logprobs, values = self.batched_forward_pass(queries, responses)
+        model_inputs = self.prepare_model_inputs(queries, responses)
+        model_inputs_names = list(model_inputs.keys())
+
+        with torch.no_grad():
+            all_logprobs, _, values, masks = self.batched_forward_pass(self.model, queries, responses, model_inputs)
+            ref_logprobs, _, _, _ = self.batched_forward_pass(self.ref_model, queries, responses, model_inputs)
+
         timing["time/ppo/forward_pass"] = time.time() - t
 
         t = time.time()
-        rewards, non_score_reward = self.compute_rewards(scores, logprobs, ref_logprobs)
+        rewards, non_score_reward = self.compute_rewards(scores, all_logprobs, ref_logprobs, masks)
         timing["time/ppo/compute_rewards"] = time.time() - t
 
+        mini_batch_dict = {
+            "queries": queries,
+            "responses": responses,
+            "logprobs": all_logprobs,
+            "values": values,
+            "rewards": rewards,
+            "masks": masks,
+        }
+
+        def collator(data):
+            return_dict = dict()
+            for key in data[0]:
+                if key in ["queries", "responses"]:
+                    return_dict[key] = [d[key] for d in data]
+                else:
+                    return_dict[key] = torch.stack([d[key] for d in data]).to(self.accelerator.device)
+            return return_dict
+
+        mini_batch_dict.update(model_inputs)
+        mini_batch_data = Dataset.from_dict(mini_batch_dict)
+        mini_batch_data.set_format("torch")
+        mini_batch_dataloader = torch.utils.data.DataLoader(
+            mini_batch_data,
+            batch_size=self.config.mini_batch_size,
+            shuffle=True,
+            collate_fn=collator,
+        )
+
         t = time.time()
         all_stats = []
-        idxs = list(range(bs))
         for _ in range(self.config.ppo_epochs):
-            random.shuffle(idxs)
-            for i in range(bs):
-                idx = idxs[i]
+            for batch in mini_batch_dataloader:
+                model_inputs = {k: batch[k] for k in model_inputs_names}
+                logprobs, logits, vpreds, _ = self.batched_forward_pass(
+                    self.model, batch["queries"], batch["responses"], model_inputs
+                )
+
                 train_stats = self.train_minibatch(
-                    logprobs[idx].unsqueeze(0),
-                    values[idx].unsqueeze(0),
-                    rewards[idx].unsqueeze(0),
-                    queries[idx].unsqueeze(0),
-                    responses[idx].unsqueeze(0),
-                    torch.cat([queries[idx], responses[idx]]).unsqueeze(0),
+                    batch["logprobs"],
+                    batch["values"],
+                    batch["rewards"],
+                    logprobs,
+                    logits,
+                    vpreds,
+                    batch["masks"],
                 )
                 all_stats.append(train_stats)
         timing["time/ppo/optimize_step"] = time.time() - t
@@ -454,11 +495,12 @@ class PPOTrainer(BaseTrainer):
 
         stats = self.record_step_stats(
             scores=scores,
-            logprobs=logprobs,
+            logprobs=all_logprobs,
             ref_logprobs=ref_logprobs,
             non_score_reward=non_score_reward,
             train_stats=train_stats,
             kl_coef=self.kl_ctl.value,
+            masks=masks,
         )
         # Gather/Reduce stats from all processes
         if self.is_distributed:
@@ -506,10 +548,34 @@ class PPOTrainer(BaseTrainer):
             stats[k] = v
         return stats
 
+    def prepare_model_inputs(self, queries: torch.Tensor, responses: torch.Tensor):
+        if self.is_encoder_decoder:
+            input_data = self.data_collator(
+                [{"input_ids": q, "attention_mask": torch.ones_like(q)} for q in queries]
+            ).to(self.accelerator.device)
+
+            decoder_inputs = self.data_collator(
+                [{"input_ids": r, "attention_mask": torch.ones_like(r)} for r in responses]
+            ).to(self.accelerator.device)
+
+            input_data["decoder_input_ids"] = decoder_inputs["input_ids"]
+            input_data["decoder_attention_mask"] = decoder_inputs["attention_mask"]
+
+            input_data.pop("labels", None)  # we don't want to compute LM losses
+
+        else:
+            input_ids = [torch.cat([q, r]) for q, r in zip(queries, responses)]
+            input_data = self.data_collator(
+                [{"input_ids": ids, "attention_mask": torch.ones_like(ids)} for ids in input_ids]
+            ).to(self.accelerator.device)
+        return input_data
+
     def batched_forward_pass(
         self,
+        model: PreTrainedModelWrapper,
         queries: torch.Tensor,
         responses: torch.Tensor,
+        model_inputs: dict,
     ):
         """
         Calculate model outputs in multiple batches.
@@ -521,74 +587,74 @@ class PPOTrainer(BaseTrainer):
                 List of tensors containing the encoded responses, shape (`batch_size`, `response_length`)
         Returns:
             (tuple):
-                - all_logprobs (`torch.FloatTensor`): Log probabilities of the responses, shape (`batch_size`, `response_length`)
-                - all_ref_logprobs (`torch.FloatTensor`): Log probabilities of the responses, shape (`batch_size`, `response_length`)
+                - all_logprobs (`torch.FloatTensor`): Log probabilities of the responses,
+                    shape (`batch_size`, `response_length`)
+                - all_ref_logprobs (`torch.FloatTensor`): Log probabilities of the responses,
+                    shape (`batch_size`, `response_length`)
                 - all_values (`torch.FloatTensor`): Values of the responses, shape (`batch_size`, `response_length`)
         """
-        bs = self.config.batch_size
-        fbs = self.config.forward_batch_size
+        bs = len(queries)
+        fbs = self.config.mini_batch_size
         all_logprobs = []
-        all_ref_logprobs = []
+        all_logits = []
+        all_masks = []
         all_values = []
 
         for i in range(int(bs / fbs)):
+            input_kwargs = {key: value[i * fbs : (i + 1) * fbs] for key, value in model_inputs.items()}
             query_batch = queries[i * fbs : (i + 1) * fbs]
             response_batch = responses[i * fbs : (i + 1) * fbs]
+            logits, _, values = model(**input_kwargs)
 
             if self.is_encoder_decoder:
-                input_ids = self.data_collator(query_batch)["input_ids"]
-                decoder_input_ids = self.data_collator(response_batch)["input_ids"]
-
-                input_kwargs = {
-                    "input_ids": input_ids,
-                    "decoder_input_ids": decoder_input_ids,
-                }
+                input_ids = input_kwargs["decoder_input_ids"]
+                attention_mask = input_kwargs["decoder_attention_mask"]
             else:
-                input_ids = self.data_collator([torch.cat([q, r]) for q, r in zip(query_batch, response_batch)])[
-                    "input_ids"
-                ]
+                input_ids = input_kwargs["input_ids"]
+                attention_mask = input_kwargs["attention_mask"]
 
-                input_kwargs = {
-                    "input_ids": input_ids,
-                }
-
-            with torch.no_grad():
-                logits, _, v = self.model(**input_kwargs)
-                ref_logits, _, _ = self.ref_model(**input_kwargs)
-
-            if self.is_encoder_decoder:
-                logprobs = logprobs_from_logits(logits[:, :-1, :], decoder_input_ids[:, 1:])
-                ref_logprobs = logprobs_from_logits(ref_logits[:, :-1, :], decoder_input_ids[:, 1:])
-            else:
-                logprobs = logprobs_from_logits(logits[:, :-1, :], input_ids[:, 1:])
-                ref_logprobs = logprobs_from_logits(ref_logits[:, :-1, :], input_ids[:, 1:])
+            logprobs = logprobs_from_logits(logits[:, :-1, :], input_ids[:, 1:])
+            masks = torch.zeros_like(attention_mask)
+            masks[:, :-1] = attention_mask[:, 1:]
 
             for j in range(fbs):
                 if self.is_encoder_decoder:
                     # Decoder sentence starts always in the index 1 after padding in the Enc-Dec Models
                     start = 1
-                    end = response_batch[j].shape[-1] - 1
+                    end = attention_mask[j, :].sum() - 1
                 else:
                     start = len(query_batch[j]) - 1
-                    end = len(query_batch[j]) + len(response_batch[j]) - 1
+                    if attention_mask[j, 0] == 0:  # offset left padding
+                        start += attention_mask[j, :].nonzero()[0]
+                    end = start + len(response_batch[j])
 
                 if len(logprobs[j, start:end]) < 2:
                     raise ValueError("Responses are too short. Make sure they are at least 4 tokens long.")
 
-                all_values.append(v[j, start - 1 : end - 1])
-                all_logprobs.append(logprobs[j, start:end])
-                all_ref_logprobs.append(ref_logprobs[j, start:end])
+                masks[j, :start] = 0
+                masks[j, end:] = 0
 
-        return all_logprobs, all_ref_logprobs, all_values
+            all_logits.append(logits)
+            all_values.append(values)
+            all_logprobs.append(logprobs)
+            all_masks.append(masks)
+
+        return (
+            torch.cat(all_logprobs),
+            torch.cat(all_logits)[:, :-1],
+            torch.cat(all_values)[:, :-1],
+            torch.cat(all_masks)[:, :-1],
+        )
 
     def train_minibatch(
         self,
-        logprobs: torch.FloatTensor,
+        old_logprobs: torch.FloatTensor,
         values: torch.FloatTensor,
         rewards: torch.FloatTensor,
-        query: torch.LongTensor,
-        response: torch.LongTensor,
-        model_input: torch.LongTensor,
+        logprobs: torch.FloatTensor,
+        logits: torch.FloatTensor,
+        vpreds: torch.FloatTensor,
+        mask: torch.LongTensor,
     ):
         """
         Train one PPO minibatch
@@ -611,16 +677,29 @@ class PPOTrainer(BaseTrainer):
             train_stats (dict[str, `torch.Tensor`]):
                 Dictionary of training statistics
         """
-        loss_p, loss_v, train_stats = self.loss(logprobs, values, rewards, query, response, model_input)
+
+        loss_p, loss_v, train_stats = self.loss(old_logprobs, values, rewards, logits, vpreds, logprobs, mask)
         loss = loss_p + loss_v
         self.optimizer.zero_grad()
         self.accelerator.backward(loss)
+
+        if self.config.max_grad_norm is not None:
+            torch.nn.utils.clip_grad_norm_(
+                filter(lambda p: p.requires_grad, self.model.parameters()), self.config.max_grad_norm
+            )
+
         t = time.time()
         self.optimizer.step()
         train_stats["time/ppo/optimizer_step"] = torch.Tensor([time.time() - t]).to(self.accelerator.device)
         return train_stats
 
-    def compute_rewards(self, scores: torch.FloatTensor, logprobs: torch.FloatTensor, ref_logprobs: torch.FloatTensor):
+    def compute_rewards(
+        self,
+        scores: torch.FloatTensor,
+        logprobs: torch.FloatTensor,
+        ref_logprobs: torch.FloatTensor,
+        masks: torch.LongTensor,
+    ):
         """
         Compute per token rewards from scores and KL-penalty.
 
@@ -633,23 +712,25 @@ class PPOTrainer(BaseTrainer):
                 Log probabilities of the reference model, shape (`batch_size`, `response_length`)
         """
         rewards, non_score_rewards = [], []
-        for score, logprob, ref_logprob in zip(scores, logprobs, ref_logprobs):
+        for score, logprob, ref_logprob, mask in zip(scores, logprobs, ref_logprobs, masks):
             kl = logprob - ref_logprob
             non_score_reward = -self.kl_ctl.value * kl
             non_score_rewards.append(non_score_reward)
             reward = non_score_reward.clone()
-            reward[-1] += score
+            last_non_masked_index = mask.nonzero()[-1]
+            reward[last_non_masked_index] += score
             rewards.append(reward)
-        return rewards, non_score_rewards
+        return torch.stack(rewards), torch.stack(non_score_rewards)
 
     def loss(
         self,
         old_logprobs: torch.FloatTensor,
         values: torch.FloatTensor,
         rewards: torch.FloatTensor,
-        query: torch.LongTensor,
-        response: torch.LongTensor,
-        model_input: torch.LongTensor,
+        logits: torch.FloatTensor,
+        vpreds: torch.FloatTensor,
+        logprobs: torch.FloatTensor,
+        mask: torch.LongTensor,
     ):
         """
         Calculate policy and value losses.
@@ -661,17 +742,20 @@ class PPOTrainer(BaseTrainer):
                 Values of the value head, shape (`batch_size`, `hidden_dim`)
             rewards (`torch.FloatTensor`):
                 Rewards from the reward model, shape (`batch_size`)
-            query (`torch.LongTensor`):
-                Encoded queries, shape (`batch_size`, `query_length`)
-            response (`torch.LongTensor`):
-                Encoded responses, shape (`batch_size`, `response_length`)
-            model_input (`torch.LongTensor`):
-                Concatenated queries and responses, shape (`batch_size`, `query_length+response_length`)
+            logits (`torch.FloatTensor`):
+                Logits of the model, shape (`batch_size`, `response_length`, `vocab_size`)
+            v_pred (`torch.FloatTensor`):
+                Values of the value head, shape (`batch_size`, `response_length`)
+            logprobs (`torch.FloatTensor`):
+                Log probabilities of the model, shape (`batch_size`, `response_length`)
         """
         lastgaelam = 0
         advantages_reversed = []
         gen_len = rewards.shape[-1]
 
+        values = values * mask
+        rewards = rewards * mask
+
         for t in reversed(range(gen_len)):
             nextvalues = values[:, t + 1] if t < gen_len - 1 else 0.0
             delta = rewards[:, t] + self.config.gamma * nextvalues - values[:, t]
@@ -680,50 +764,32 @@ class PPOTrainer(BaseTrainer):
         advantages = torch.stack(advantages_reversed[::-1]).transpose(0, 1)
 
         returns = advantages + values
-        advantages = whiten(advantages)
+        advantages = masked_whiten(advantages, mask)
         advantages = advantages.detach()
 
-        input_kwargs = {
-            "input_ids": model_input,
-        }
+        vpredclipped = clip_by_value(
+            vpreds, values - self.config.cliprange_value, values + self.config.cliprange_value
+        )
 
-        if self.is_encoder_decoder:
-            input_kwargs["input_ids"] = query
-            input_kwargs["decoder_input_ids"] = response
-            model_input = response
-
-        logits, _, vpred = self.model(**input_kwargs)
-
-        if self.is_encoder_decoder:
-            logprob = logprobs_from_logits(logits[:, :-1, :], model_input[:, 1:])
-            start, end = 1, response.shape[-1] - 1
-            vpred = vpred[:, start - 1 : end - 1]
-            logprob = logprob[:, start:end]
-        else:
-            logprob = logprobs_from_logits(logits[:, :-1, :], model_input[:, 1:])
-            logprob, vpred = logprob[:, -gen_len:], vpred[:, -gen_len - 1 : -1]
-
-        vpredclipped = clip_by_value(vpred, values - self.config.cliprange_value, values + self.config.cliprange_value)
-
-        vf_losses1 = (vpred - returns) ** 2
+        vf_losses1 = (vpreds - returns) ** 2
         vf_losses2 = (vpredclipped - returns) ** 2
-        vf_loss = 0.5 * torch.mean(torch.max(vf_losses1, vf_losses2))
-        vf_clipfrac = torch.mean(torch.gt(vf_losses2, vf_losses1).double())
+        vf_loss = 0.5 * masked_mean(torch.max(vf_losses1, vf_losses2), mask)
+        vf_clipfrac = masked_mean(torch.gt(vf_losses2, vf_losses1).double(), mask)
 
-        ratio = torch.exp(logprob - old_logprobs)
+        ratio = torch.exp(logprobs - old_logprobs)
         pg_losses = -advantages * ratio
         pg_losses2 = -advantages * torch.clamp(ratio, 1.0 - self.config.cliprange, 1.0 + self.config.cliprange)
 
-        pg_loss = torch.mean(torch.max(pg_losses, pg_losses2))
-        pg_clipfrac = torch.mean(torch.gt(pg_losses2, pg_losses).double())
+        pg_loss = masked_mean(torch.max(pg_losses, pg_losses2), mask)
+        pg_clipfrac = masked_mean(torch.gt(pg_losses2, pg_losses).double(), mask)
 
         loss = pg_loss + self.config.vf_coef * vf_loss
 
-        entropy = torch.mean(entropy_from_logits(logits))
-        approxkl = 0.5 * torch.mean((logprob - old_logprobs) ** 2)
-        policykl = torch.mean(logprob - old_logprobs)
-        return_mean, return_var = torch.mean(returns), torch.var(returns)
-        value_mean, value_var = torch.mean(values), torch.var(values)
+        entropy = masked_mean(entropy_from_logits(logits), mask)
+        approxkl = 0.5 * masked_mean((logprobs - old_logprobs) ** 2, mask)
+        policykl = masked_mean(old_logprobs - logprobs, mask)
+        return_mean, return_var = masked_mean(returns, mask), masked_var(returns, mask)
+        value_mean, value_var = masked_mean(values, mask), masked_var(values, mask)
 
         stats = dict(
             loss=dict(policy=pg_loss, value=vf_loss, total=loss),
@@ -733,13 +799,13 @@ class PPOTrainer(BaseTrainer):
                 policykl=policykl,
                 clipfrac=pg_clipfrac,
                 advantages=advantages,
-                advantages_mean=torch.mean(advantages),
+                advantages_mean=masked_mean(advantages, mask),
                 ratio=ratio,
             ),
             returns=dict(mean=return_mean, var=return_var),
             val=dict(
-                vpred=torch.mean(vpred),
-                error=torch.mean((vpred - returns) ** 2),
+                vpred=masked_mean(vpreds, mask),
+                error=masked_mean((vpreds - returns) ** 2, mask),
                 clipfrac=vf_clipfrac,
                 mean=value_mean,
                 var=value_var,
@@ -762,12 +828,14 @@ class PPOTrainer(BaseTrainer):
             stats (`dict`):
                 Dictionary of training step statistics
         """
-        kl_list = [logprobs - ref_logprobs for logprobs, ref_logprobs in zip(data["logprobs"], data["ref_logprobs"])]
-        mean_kl = torch.mean(torch.stack([torch.sum(kl) for kl in kl_list]))
-        mean_entropy = torch.mean(torch.stack([torch.sum(-log_probs) for log_probs in data["logprobs"]]))
-        mean_non_score_reward = torch.mean(
-            torch.stack([torch.sum(non_score_reward) for non_score_reward in data["non_score_reward"]])
-        )
+        mask = data.pop("masks")
+
+        kl_list = ((data["logprobs"] - data["ref_logprobs"]) * mask).sum(axis=-1)
+        mean_kl = kl_list.mean()
+        mean_entropy = (-data["logprobs"] * mask).sum(axis=-1).mean()
+
+        mean_non_score_reward = masked_mean(data["non_score_reward"], mask)
+
         stats = {
             "objective/kl": mean_kl,
             "objective/kl_dist": kl_list,
@@ -796,7 +864,7 @@ class PPOTrainer(BaseTrainer):
             stats (dict[str, Any]):
                 A dictionary of training stats.
             batch (dict[str, Any]):
-                A dictionary of batch data, this containes the queries and responses.
+                A dictionary of batch data, this contains the queries and responses.
             rewards (`List[torch.FloatTensor]`):
                 A tensor of rewards.
         """
@@ -811,7 +879,8 @@ class PPOTrainer(BaseTrainer):
             if "query" not in batch.keys() and "response" not in batch.keys():
                 # warn the user that the game logs will not be logged
                 warnings.warn(
-                    "The game logs will not be logged because the batch does not contain the keys 'query' and 'response'."
+                    "The game logs will not be logged because the batch does not contain the keys 'query' and "
+                    "'response'. "
                 )
             elif self.config.log_with == "wandb":
                 import wandb
@@ -828,6 +897,12 @@ class PPOTrainer(BaseTrainer):
                 rewards /= self.accelerator.num_processes
 
             logs.update(stats)
+
+            # manually cast in fp32 for bf16 torch tensors
+            for k, v in logs.items():
+                if isinstance(v, torch.Tensor) and v.dtype == torch.bfloat16:
+                    logs[k] = v.float()
+
             logs["env/reward_mean"] = torch.mean(rewards).cpu().numpy().item()
             logs["env/reward_std"] = torch.std(rewards).cpu().numpy().item()
             logs["env/reward_dist"] = rewards.cpu().numpy()
@@ -864,6 +939,6 @@ class PPOTrainer(BaseTrainer):
             f.write(model_card_content)
 
     def _save_pretrained(self, save_directory: str) -> None:
-        self.model.save_pretrained(save_directory)
+        self.accelerator.unwrap_model(self.model).save_pretrained(save_directory)
         self.tokenizer.save_pretrained(save_directory)
         self.create_model_card(save_directory)
